@@ -34,6 +34,10 @@
 *     [13] Cortes, David.
 *          "Isolation forests: looking beyond tree depth."
 *          arXiv preprint arXiv:2111.11639 (2021).
+*     [14] Ting, Kai Ming, Yue Zhu, and Zhi-Hua Zhou.
+*          "Isolation kernel and its effect on SVM"
+*          Proceedings of the 24th ACM SIGKDD
+*          International Conference on Knowledge Discovery & Data Mining. 2018.
 * 
 *     BSD 2-Clause License
 *     Copyright (c) 2019-2022, David Cortes
@@ -195,6 +199,8 @@ void predict_iforest(real_t *restrict numeric_data, int *restrict categ_data,
                      double *restrict per_tree_depths,
                      TreesIndexer *indexer)
 {
+    if (unlikely(!nrows)) return;
+
     /* put data in a struct for passing it in fewer lines */
     PredictionData<real_t, sparse_ix>
                    prediction_data = {numeric_data, categ_data, nrows,
@@ -220,7 +226,7 @@ void predict_iforest(real_t *restrict numeric_data, int *restrict categ_data,
     {
         if (
             model_outputs->missing_action == Fail &&
-            (model_outputs->new_cat_action != Weighted || prediction_data.categ_data == NULL) &&
+            (model_outputs->new_cat_action != Weighted || model_outputs->cat_split_type == SingleCateg || prediction_data.categ_data == NULL) &&
             prediction_data.Xc_indptr == NULL && prediction_data.Xr_indptr == NULL &&
             !model_outputs->has_range_penalty
             )
@@ -435,7 +441,7 @@ void predict_iforest(real_t *restrict numeric_data, int *restrict categ_data,
             {
                 if (model_outputs->missing_action == Divide)
                     goto manual_remap;
-                if (model_outputs->new_cat_action == Weighted && categ_data != NULL)
+                if (model_outputs->new_cat_action == Weighted && model_outputs->cat_split_type == SubSet && categ_data != NULL)
                     goto manual_remap;
             }
 
@@ -465,7 +471,7 @@ void traverse_itree_no_recurse(std::vector<IsoTree>  &tree,
                                double &restrict      output_depth,
                                sparse_ix *restrict   tree_num,
                                double *restrict      tree_depth,
-                               size_t                row)
+                               size_t                row) noexcept
 {
     size_t curr_lev = 0;
     double xval;
@@ -473,12 +479,12 @@ void traverse_itree_no_recurse(std::vector<IsoTree>  &tree,
     while (true)
     {
         // if (tree[curr_lev].score > 0)
-        if (tree[curr_lev].tree_left == 0)
+        if (unlikely(tree[curr_lev].tree_left == 0))
         {
             output_depth += tree[curr_lev].score;
-            if (tree_num != NULL)
+            if (unlikely(tree_num != NULL))
                 tree_num[row] = curr_lev;
-            if (tree_depth != NULL)
+            if (unlikely(tree_depth != NULL))
                 *tree_depth = tree[curr_lev].score;
             break;
         }
@@ -543,7 +549,7 @@ void traverse_itree_no_recurse(std::vector<IsoTree>  &tree,
 
                                     case Smallest:
                                     {
-                                        if (cval >= (int)tree[curr_lev].cat_split.size())
+                                        if (unlikely(cval >= (int)tree[curr_lev].cat_split.size()))
                                         {
                                             curr_lev =  (tree[curr_lev].pct_tree_left < .5)? tree[curr_lev].tree_left : tree[curr_lev].tree_right;
                                         }
@@ -598,7 +604,7 @@ double traverse_itree(std::vector<IsoTree>     &tree,
                       size_t                   row,
                       sparse_ix *restrict      tree_num,
                       double *restrict         tree_depth,
-                      size_t                   curr_lev)
+                      size_t                   curr_lev) noexcept
 {
     double xval;
     int    cval;
@@ -624,13 +630,13 @@ double traverse_itree(std::vector<IsoTree>     &tree,
     while (true)
     {
         // if (tree[curr_lev].score >= 0.)
-        if (tree[curr_lev].tree_left == 0)
+        if (unlikely(tree[curr_lev].tree_left == 0))
         {
-            if (tree_num != NULL)
+            if (unlikely(tree_num != NULL))
                 tree_num[row] = curr_lev;
-            if (tree_depth != NULL)
+            if (unlikely(tree_depth != NULL))
                 *tree_depth = tree[curr_lev].score;
-            if (imputed_data != NULL)
+            if (unlikely(imputed_data != NULL))
                 add_from_impute_node((*impute_nodes)[curr_lev], *imputed_data, curr_weight);
 
             return tree[curr_lev].score - range_penalty;
@@ -669,7 +675,7 @@ double traverse_itree(std::vector<IsoTree>     &tree,
                         }
                     }
 
-                    if (isnan(xval))
+                    if (unlikely(isnan(xval)))
                     {
                         switch(model_outputs.missing_action)
                         {
@@ -718,7 +724,7 @@ double traverse_itree(std::vector<IsoTree>     &tree,
                                     :
                                 (tree[curr_lev].col_num + row * prediction_data.ncols_categ)
                             ];
-                    if (cval < 0)
+                    if (unlikely(cval < 0))
                     {
                         switch(model_outputs.missing_action)
                         {
@@ -820,7 +826,7 @@ double traverse_itree(std::vector<IsoTree>     &tree,
 
                                         case Smallest:
                                         {
-                                            if (cval >= (int)tree[curr_lev].cat_split.size())
+                                            if (unlikely(cval >= (int)tree[curr_lev].cat_split.size()))
                                             {
                                                 curr_lev =  (tree[curr_lev].pct_tree_left < .5)? tree[curr_lev].tree_left : tree[curr_lev].tree_right;
                                             }
@@ -886,7 +892,7 @@ void traverse_hplane_fast(std::vector<IsoHPlane>  &hplane,
                           double &restrict        output_depth,
                           sparse_ix *restrict     tree_num,
                           double *restrict        tree_depth,
-                          size_t                  row)
+                          size_t                  row) noexcept
 {
     size_t  curr_lev = 0;
     double  hval;
@@ -894,12 +900,12 @@ void traverse_hplane_fast(std::vector<IsoHPlane>  &hplane,
     while(true)
     {
         // if (hplane[curr_lev].score > 0)
-        if (hplane[curr_lev].hplane_left == 0)
+        if (unlikely(hplane[curr_lev].hplane_left == 0))
         {
             output_depth += hplane[curr_lev].score;
-            if (tree_num != NULL)
+            if (unlikely(tree_num != NULL))
                 tree_num[row] = curr_lev;
-            if (tree_depth != NULL)
+            if (unlikely(tree_depth != NULL))
                 *tree_depth = hplane[curr_lev].score;
             return;
         }
@@ -937,7 +943,7 @@ void traverse_hplane(std::vector<IsoHPlane>   &hplane,
                      ImputedData             *imputed_data,     /* only when imputing missing */
                      sparse_ix *restrict      tree_num,
                      double *restrict         tree_depth,
-                     size_t                   row)
+                     size_t                   row) noexcept
 {
     size_t  curr_lev = 0;
     double  xval;
@@ -966,14 +972,14 @@ void traverse_hplane(std::vector<IsoHPlane>   &hplane,
     while(true)
     {
         // if (hplane[curr_lev].score > 0)
-        if (hplane[curr_lev].hplane_left == 0)
+        if (unlikely(hplane[curr_lev].hplane_left == 0))
         {
             output_depth += hplane[curr_lev].score;
-            if (tree_num != NULL)
+            if (unlikely(tree_num != NULL))
                 tree_num[row] = curr_lev;
-            if (tree_depth != NULL)
+            if (unlikely(tree_depth != NULL))
                 *tree_depth = hplane[curr_lev].score;
-            if (imputed_data != NULL)
+            if (unlikely(imputed_data != NULL))
             {
                 add_from_impute_node((*impute_nodes)[curr_lev], *imputed_data, (double)1);
             }
@@ -1017,7 +1023,7 @@ void traverse_hplane(std::vector<IsoHPlane>   &hplane,
                             }
                         }
 
-                        if (is_na_or_inf(xval))
+                        if (unlikely(is_na_or_inf(xval)))
                         {
                             if (model_outputs.missing_action != Fail)
                             {
@@ -1048,7 +1054,7 @@ void traverse_hplane(std::vector<IsoHPlane>   &hplane,
                                 :
                             (hplane[curr_lev].col_num[col] + row * prediction_data.ncols_categ)
                         ];
-                        if (cval < 0)
+                        if (unlikely(cval < 0))
                         {
                             if (model_outputs.missing_action != Fail)
                             {
@@ -1074,7 +1080,7 @@ void traverse_hplane(std::vector<IsoHPlane>   &hplane,
 
                                 case SubSet:
                                 {
-                                    if (cval >= (int)hplane[curr_lev].cat_coef[ncols_categ].size())
+                                    if (unlikely(cval >= (int)hplane[curr_lev].cat_coef[ncols_categ].size()))
                                     {
                                         if (model_outputs.new_cat_action == Random) {
                                             cval = cval % (int)hplane[curr_lev].cat_coef[ncols_categ].size();
@@ -1154,7 +1160,7 @@ void batched_csc_predict(PredictionData<real_t, sparse_ix> &prediction_data, int
                               (size_t)0);
 
                     if (model_outputs->missing_action == Divide ||
-                        (model_outputs->new_cat_action == Weighted && prediction_data.categ_data != NULL)
+                        (model_outputs->new_cat_action == Weighted && model_outputs->cat_split_type == SubSet && prediction_data.categ_data != NULL)
                     ) {
                         ptr_worker->weights_arr.resize(prediction_data.nrows);
                     }
@@ -1276,7 +1282,7 @@ void traverse_itree_csc(WorkerForPredictCSC   &workspace,
                         bool                  has_range_penalty)
 {
     // if (trees[curr_tree].score >= 0)
-    if (trees[curr_tree].tree_left == 0)
+    if (unlikely(trees[curr_tree].tree_left == 0))
     {
         if (model_outputs.missing_action != Divide)
             for (size_t row = workspace.st; row <= workspace.end; row++)
@@ -1284,10 +1290,10 @@ void traverse_itree_csc(WorkerForPredictCSC   &workspace,
         else
             for (size_t row = workspace.st; row <= workspace.end; row++)
                 workspace.depths[workspace.ix_arr[row]] += workspace.weights_arr[workspace.ix_arr[row]] * trees[curr_tree].score;
-        if (tree_num != NULL)
+        if (unlikely(tree_num != NULL))
             for (size_t row = workspace.st; row <= workspace.end; row++)
                 tree_num[workspace.ix_arr[row]] = curr_tree;
-        if (per_tree_depths != NULL)
+        if (unlikely(per_tree_depths != NULL))
             for (size_t row = workspace.st; row <= workspace.end; row++)
                 per_tree_depths[workspace.ix_arr[row]] = trees[curr_tree].score;
         return;
@@ -1302,7 +1308,7 @@ void traverse_itree_csc(WorkerForPredictCSC   &workspace,
     /* divide according to tree */
     size_t orig_end = workspace.end;
     size_t st_NA, end_NA, split_ix;
-    switch(trees[curr_tree].col_type)
+    switch (trees[curr_tree].col_type)
     {
         case Numeric:
         {
@@ -1315,7 +1321,7 @@ void traverse_itree_csc(WorkerForPredictCSC   &workspace,
 
         case Categorical:
         {
-            switch(model_outputs.cat_split_type)
+            switch (model_outputs.cat_split_type)
             {
                 case SingleCateg:
                 {
@@ -1355,9 +1361,9 @@ void traverse_itree_csc(WorkerForPredictCSC   &workspace,
     }
 
     /* continue splitting recursively */
-    if (model_outputs.new_cat_action == Weighted)
+    if (unlikely(model_outputs.new_cat_action == Weighted && model_outputs.cat_split_type == SubSet && prediction_data.categ_data != NULL))
         goto missing_action_divide;
-    switch(model_outputs.missing_action)
+    switch (model_outputs.missing_action)
     {
         case Impute:
         {
@@ -1521,14 +1527,14 @@ void traverse_hplane_csc(WorkerForPredictCSC      &workspace,
                          bool                     has_range_penalty)
 {
     // if (hplanes[curr_tree].score >= 0)
-    if (hplanes[curr_tree].hplane_left == 0)
+    if (unlikely(hplanes[curr_tree].hplane_left == 0))
     {
         for (size_t row = workspace.st; row <= workspace.end; row++)
             workspace.depths[workspace.ix_arr[row]] += hplanes[curr_tree].score;
-        if (tree_num != NULL)
+        if (unlikely(tree_num != NULL))
             for (size_t row = workspace.st; row <= workspace.end; row++)
                 tree_num[workspace.ix_arr[row]] = curr_tree;
-        if (per_tree_depths != NULL)
+        if (unlikely(per_tree_depths != NULL))
             for (size_t row = workspace.st; row <= workspace.end; row++)
                 per_tree_depths[workspace.ix_arr[row]] = hplanes[curr_tree].score;
         return;
@@ -1538,7 +1544,7 @@ void traverse_hplane_csc(WorkerForPredictCSC      &workspace,
     std::fill(workspace.comb_val.begin(), workspace.comb_val.begin() + (workspace.end - workspace.st + 1), 0.);
     double unused;
 
-    if (prediction_data.categ_data == NULL)
+    if (likely(prediction_data.categ_data == NULL))
     {
         for (size_t col = 0; col < hplanes[curr_tree].col_num.size(); col++)
             add_linear_comb(workspace.ix_arr.data(), workspace.st, workspace.end,
@@ -1555,7 +1561,7 @@ void traverse_hplane_csc(WorkerForPredictCSC      &workspace,
         size_t ncols_categ = 0;
         for (size_t col = 0; col < hplanes[curr_tree].col_num.size(); col++)
         {
-            switch(hplanes[curr_tree].col_type[col])
+            switch (hplanes[curr_tree].col_type[col])
             {
                 case Numeric:
                 {
@@ -1661,10 +1667,10 @@ void add_csc_range_penalty(WorkerForPredictCSC  &workspace,
         {
             if (prediction_data.Xc_ind[curr_pos] == (decltype(*prediction_data.Xc_ind))(*row))
             {
-                if (!isnan(prediction_data.Xc[curr_pos])
-                        &&
-                    (   prediction_data.Xc[curr_pos] < range_low    ||
-                        prediction_data.Xc[curr_pos] > range_high   ))
+                if (likely(!isnan(prediction_data.Xc[curr_pos])
+                               &&
+                           (   prediction_data.Xc[curr_pos] < range_low    ||
+                               prediction_data.Xc[curr_pos] > range_high   )))
                 {
                     workspace.depths[*row] -= (weights_arr == NULL)? 1. : weights_arr[*row];
                 }
@@ -1693,7 +1699,7 @@ void add_csc_range_penalty(WorkerForPredictCSC  &workspace,
 
     else
     {
-        if (weights_arr == NULL)
+        if (likely(weights_arr == NULL))
             for (size_t row = workspace.st; row <= workspace.end; row++)
                 workspace.depths[workspace.ix_arr[row]]--;
         else
@@ -1707,10 +1713,10 @@ void add_csc_range_penalty(WorkerForPredictCSC  &workspace,
         {
             if (prediction_data.Xc_ind[curr_pos] == (decltype(*prediction_data.Xc_ind))(*row))
             {
-                if (isnan(prediction_data.Xc[curr_pos])
-                        ||
-                    (   prediction_data.Xc[curr_pos] >= range_low    &&
-                        prediction_data.Xc[curr_pos] <= range_high   ))
+                if (likely(isnan(prediction_data.Xc[curr_pos])
+                               ||
+                           (   prediction_data.Xc[curr_pos] >= range_low    &&
+                               prediction_data.Xc[curr_pos] <= range_high   )))
                 {
                     workspace.depths[*row] += (weights_arr == NULL)? 1. : weights_arr[*row];
                 }
@@ -1739,7 +1745,7 @@ void add_csc_range_penalty(WorkerForPredictCSC  &workspace,
 }
 
 template <class PredictionData>
-double extract_spC(PredictionData &prediction_data, size_t row, size_t col_num)
+double extract_spC(PredictionData &prediction_data, size_t row, size_t col_num) noexcept
 {
     decltype(prediction_data.Xc_indptr)
                search_res = std::lower_bound(prediction_data.Xc_ind + prediction_data.Xc_indptr[col_num],
@@ -1756,7 +1762,7 @@ double extract_spC(PredictionData &prediction_data, size_t row, size_t col_num)
 }
 
 template <class PredictionData, class sparse_ix>
-double extract_spR(PredictionData &prediction_data, sparse_ix *row_st, sparse_ix *row_end, size_t col_num)
+double extract_spR(PredictionData &prediction_data, sparse_ix *row_st, sparse_ix *row_end, size_t col_num) noexcept
 {
     if (row_end == row_st)
         return 0.;
@@ -1768,7 +1774,7 @@ double extract_spR(PredictionData &prediction_data, sparse_ix *row_st, sparse_ix
 }
 
 template <class sparse_ix>
-void get_num_nodes(IsoForest &model_outputs, sparse_ix *restrict n_nodes, sparse_ix *restrict n_terminal, int nthreads)
+void get_num_nodes(IsoForest &model_outputs, sparse_ix *restrict n_nodes, sparse_ix *restrict n_terminal, int nthreads) noexcept
 {
     std::fill(n_terminal, n_terminal + model_outputs.trees.size(), 0);
     #pragma omp parallel for schedule(static) num_threads(nthreads) shared(model_outputs, n_nodes, n_terminal)
@@ -1783,7 +1789,7 @@ void get_num_nodes(IsoForest &model_outputs, sparse_ix *restrict n_nodes, sparse
 }
 
 template <class sparse_ix>
-void get_num_nodes(ExtIsoForest &model_outputs, sparse_ix *restrict n_nodes, sparse_ix *restrict n_terminal, int nthreads)
+void get_num_nodes(ExtIsoForest &model_outputs, sparse_ix *restrict n_nodes, sparse_ix *restrict n_terminal, int nthreads) noexcept
 {
     std::fill(n_terminal, n_terminal + model_outputs.hplanes.size(), 0);
     #pragma omp parallel for schedule(static) num_threads(nthreads) shared(model_outputs, n_nodes, n_terminal)
@@ -1796,4 +1802,3 @@ void get_num_nodes(ExtIsoForest &model_outputs, sparse_ix *restrict n_nodes, spa
         }
     }
 }
-
