@@ -18,22 +18,29 @@ using std::uint32_t;
 using std::uint64_t;
 using std::memcpy;
 
+#ifndef _FOR_R
+    #if defined(__clang__)
+        #pragma clang diagnostic push
+        #pragma clang diagnostic ignored "-Wunknown-attributes"
+    #endif
+#endif
+
 namespace Xoshiro {
 
 #if (__cplusplus  >= 202002L)
 #define rotl64(x, k) std::rotl(x, k)
 #define rotl32(x, k) std::rotl(x, k)
 #else
-static inline uint64_t rotl64(const uint64_t x, const int k) {
+static inline uint64_t rotl64(const uint64_t x, const int k) noexcept {
     return (x << k) | (x >> (64 - k));
 }
-static inline uint32_t rotl32(const uint32_t x, const int k) {
+static inline uint32_t rotl32(const uint32_t x, const int k) noexcept {
     return (x << k) | (x >> (32 - k));
 }
 #endif
 
 /* these are in order to avoid gcc warnings about 'strict aliasing rules' */
-static inline uint32_t extract_32bits_from64_left(const uint64_t x)
+static inline uint32_t extract_32bits_from64_left(const uint64_t x) noexcept
 {
     uint32_t out;
     memcpy(reinterpret_cast<char*>(&out),
@@ -42,7 +49,7 @@ static inline uint32_t extract_32bits_from64_left(const uint64_t x)
     return out;
 }
 
-static inline uint32_t extract_32bits_from64_right(const uint64_t x)
+static inline uint32_t extract_32bits_from64_right(const uint64_t x) noexcept
 {
     uint32_t out;
     memcpy(reinterpret_cast<char*>(&out),
@@ -51,14 +58,14 @@ static inline uint32_t extract_32bits_from64_right(const uint64_t x)
     return out;
 }
 
-static inline void assign_32bits_to64_left(uint64_t &assign_to, const uint32_t take_from)
+static inline void assign_32bits_to64_left(uint64_t &assign_to, const uint32_t take_from) noexcept
 {
     memcpy(reinterpret_cast<char*>(&assign_to),
            reinterpret_cast<const char*>(&take_from),
            sizeof(uint32_t));
 }
 
-static inline void assign_32bits_to64_right(uint64_t &assign_to, const uint32_t take_from)
+static inline void assign_32bits_to64_right(uint64_t &assign_to, const uint32_t take_from) noexcept
 {
     memcpy(reinterpret_cast<char*>(&assign_to) + sizeof(uint32_t),
            reinterpret_cast<const char*>(&take_from),
@@ -71,7 +78,7 @@ static inline void assign_32bits_to64_right(uint64_t &assign_to, const uint32_t
 
    It is a very fast generator passing BigCrush, and it can be useful if
    for some reason you absolutely want 64 bits of state. */
-static inline uint64_t splitmix64(const uint64_t seed)
+static inline uint64_t splitmix64(const uint64_t seed) noexcept
 {
     uint64_t z = (seed + 0x9e3779b97f4a7c15);
     z = (z ^ (z >> 30)) * 0xbf58476d1ce4e5b9;
@@ -95,19 +102,19 @@ public:
     using result_type = uint64_t;
     uint64_t state[4];
 
-    constexpr static result_type min()
+    constexpr static result_type min() noexcept
     {
         return 0;
     }
 
-    constexpr static result_type max()
+    constexpr static result_type max() noexcept
     {
         return UINT64_MAX;
     }
 
-    Xoshiro256PP() = default;
+    Xoshiro256PP() noexcept = default;
 
-    inline void seed(const uint64_t seed)
+    inline void seed(const uint64_t seed) noexcept
     {
         this->state[0] = splitmix64(splitmix64(seed));
         this->state[1] = splitmix64(this->state[0]);
@@ -116,23 +123,23 @@ public:
     }
 
     template <class integer>
-    inline void seed(const integer seed)
+    inline void seed(const integer seed) noexcept
     {
         this->seed((uint64_t)seed);
     }
 
-    Xoshiro256PP(const uint64_t seed)
+    Xoshiro256PP(const uint64_t seed) noexcept
     {
         this->seed(seed);
     }
 
     template <class integer>
-    Xoshiro256PP(const integer seed)
+    Xoshiro256PP(const integer seed) noexcept
     {
         this->seed((uint64_t)seed);
     }
 
-    inline result_type operator()()
+    inline result_type operator()() noexcept
     {
         const uint64_t result = rotl64(this->state[0] + this->state[3], 23) + this->state[0];
         const uint64_t t = this->state[1] << 17;
@@ -161,20 +168,20 @@ public:
     using result_type = uint32_t;
     uint32_t state[4];
 
-    constexpr static result_type min()
+    constexpr static result_type min() noexcept
     {
         return 0;
     }
 
-    constexpr static result_type max()
+    constexpr static result_type max() noexcept
     {
         return UINT32_MAX;
     }
 
-    Xoshiro128PP() = default;
+    Xoshiro128PP() noexcept = default;
 
 
-    inline void seed(const uint64_t seed)
+    inline void seed(const uint64_t seed) noexcept
     {
         const auto t1 = splitmix64(seed);
         const auto t2 = splitmix64(t1);
@@ -184,7 +191,7 @@ public:
         this->state[3] = extract_32bits_from64_right(t2);
     }
 
-    inline void seed(const uint32_t seed)
+    inline void seed(const uint32_t seed) noexcept
     {
         uint64_t temp;
         assign_32bits_to64_left(temp, seed);
@@ -194,28 +201,28 @@ public:
 
 
     template <class integer>
-    inline void seed(const integer seed)
+    inline void seed(const integer seed) noexcept
     {
         this->seed((uint64_t)seed);
     }
 
-    Xoshiro128PP(const uint32_t seed)
+    Xoshiro128PP(const uint32_t seed) noexcept
     {
         this->seed(seed);
     }
 
-    Xoshiro128PP(const uint64_t seed)
+    Xoshiro128PP(const uint64_t seed) noexcept
     {
         this->seed(seed);
     }
 
     template <class integer>
-    Xoshiro128PP(const integer seed)
+    Xoshiro128PP(const integer seed) noexcept
     {
         this->seed((uint64_t)seed);
     }
 
-    inline result_type operator()()
+    inline result_type operator()() noexcept
     {
         const uint32_t result = rotl32(this->state[0] + this->state[3], 7) + this->state[0];
         const uint32_t t = this->state[1] << 9;
@@ -244,12 +251,12 @@ constexpr static const double ui64_d = (double)UINT64_MAX;
 constexpr static const double i64_d = (double)INT64_MAX;
 constexpr static const double twoPI = 2. * M_PI;
 
-static inline uint64_t gen_bits(Xoshiro256PP &rng)
+static inline uint64_t gen_bits(Xoshiro256PP &rng) noexcept
 {
     return rng();
 }
 
-static inline uint64_t gen_bits(Xoshiro128PP &rng)
+static inline uint64_t gen_bits(Xoshiro128PP &rng) noexcept
 {
     uint64_t bits;
     assign_32bits_to64_left(bits, rng());
@@ -263,7 +270,7 @@ static inline uint64_t gen_bits(Xoshiro128PP &rng)
    any further as GCC9 has a bug in which it optimizes away some 'if's'
    but with the *wrong* bit ending if done as ternary operators or if
    declaring pointer variables outside of the braces in what comes below. */
-static inline bool get_is_little_endian()
+static inline bool get_is_little_endian() noexcept
 {
     const uint32_t ONE = 1;
     return (*(reinterpret_cast<const unsigned char*>(&ONE)) != 0);
@@ -283,16 +290,16 @@ static const bool is_little_endian = get_is_little_endian();
 class UniformUnitInterval
 {
 public:
-    UniformUnitInterval() = default;
+    UniformUnitInterval() noexcept = default;
 
     template <class A, class B>
-    UniformUnitInterval(A a, B b) {}
+    UniformUnitInterval(A a, B b) noexcept {}
     
     template <class XoshiroRNG>
     #ifndef _FOR_R
     [[gnu::optimize("no-trapping-math"), gnu::optimize("no-math-errno")]]
     #endif
-    double operator()(XoshiroRNG &rng)
+    double operator()(XoshiroRNG &rng) noexcept
     {
         #if SIZE_MAX >= UINT64_MAX
         #   if (__cplusplus >= 201703L) || (__cplusplus >= 201402L && (defined(__GNUC__) || defined(_MSC_VER)))
@@ -330,16 +337,16 @@ public:
 class UniformMinusOneToOne
 {
 public:
-    UniformMinusOneToOne() = default;
+    UniformMinusOneToOne() noexcept = default;
 
     template <class A, class B>
-    UniformMinusOneToOne(A a, B b) {}
+    UniformMinusOneToOne(A a, B b) noexcept {}
 
     template <class XoshiroRNG>
     #ifndef _FOR_R
     [[gnu::optimize("no-trapping-math"), gnu::optimize("no-math-errno")]]
     #endif
-    double operator()(XoshiroRNG &rng)
+    double operator()(XoshiroRNG &rng) noexcept
     {
         #if SIZE_MAX >= UINT64_MAX
         #   if (__cplusplus >= 201703L) || (__cplusplus >= 201402L && (defined(__GNUC__) || defined(_MSC_VER)))
@@ -347,7 +354,7 @@ public:
         #   else
         double out = std::ldexp((int64_t)(gen_bits(rng) & two54_i) - two53_ii, -53);
         #endif
-        if (out == 0) out = 1;
+        if (unlikely(out == 0)) out = 1;
         return out;
         #else
         uint64_t bits = gen_bits(rng);
@@ -362,7 +369,7 @@ public:
         #   else
         double out = std::ldexp((int64_t)bits - two53_ii, -53);
         #endif
-        if (out == 0) out = 1;
+        if (unlikely(out == 0)) out = 1;
         return out;
         #endif
     }
@@ -385,16 +392,16 @@ public:
     double reserve;
     double has_reserve = false;
 
-    StandardNormalDistr() = default;
+    StandardNormalDistr() noexcept = default;
 
     template <class A, class B>
-    StandardNormalDistr(A a, B b) : has_reserve(false) {}
+    StandardNormalDistr(A a, B b) noexcept : has_reserve(false) {}
 
     template <class XoshiroRNG>
     #ifndef _FOR_R
     [[gnu::optimize("no-trapping-math"), gnu::optimize("no-math-errno")]]
     #endif
-    double operator()(XoshiroRNG &rng)
+    double operator()(XoshiroRNG &rng) noexcept
     {
         double res;
         if (has_reserve) {
@@ -448,3 +455,9 @@ public:
 };
 
 }
+
+#ifndef _FOR_R
+    #if defined(__clang__)
+        #pragma clang diagnostic pop
+    #endif
+#endif
